/* shm.cc: Single unix specification IPC interface for Cygwin

   Copyright 2001 Red Hat, Inc.

   Originally written by Robert Collins <robert.collins@hotmail.com>

   This file is part of Cygwin.

   This software is a copyrighted work licensed under the terms of the
   Cygwin license.  Please consult the file "CYGWIN_LICENSE" for
   details. */

#include "winsup.h"
#include <sys/stat.h>
#include <errno.h>
#include "cygerrno.h"
#include <unistd.h>
#include "security.h"
#include "fhandler.h"
#include "dtable.h"
#include "cygheap.h"
#include <stdio.h>
#include "thread.h"
#include <sys/shm.h>
#include "perprocess.h"
#include "cygserver_shm.h"

// FIXME IS THIS CORRECT
/* Implementation notes: We use two shared memory regions per key:
 * One for the control structure, and one for the shared memory.
 * While this has a higher overhead tham a single shared area,
 * It allows more flexability. As the entire code is transparent to the user
 * We can merge these in the future should it be needed.
 */
extern "C" size_t
getsystemallocgranularity ()
{
  SYSTEM_INFO sysinfo;
  static size_t buffer_offset = 0;
  if (buffer_offset)
    return buffer_offset;
  GetSystemInfo (&sysinfo);
  buffer_offset = sysinfo.dwAllocationGranularity;
  return buffer_offset;
}

/* FIXME: after fork, every memory area needs to have the attach count incremented
 */

extern "C" void *
shmat (int shmid, const void *shmaddr, int shmflg)
{
  class shmid_ds *shm = (class shmid_ds *) shmid;	//FIXME: verifyable object test

  if (shmaddr)
    {
      //FIXME: requested base address ?!
      set_errno (EINVAL);
      return (void *) -1;
    }

  void *rv =
    MapViewOfFile (shm->attachmap,
		   (shmflg & SHM_RDONLY) ? FILE_MAP_READ : FILE_MAP_WRITE, 0,
		   0, 0);

  if (!rv)
    {
      //FIXME: translate GetLastError()
      set_errno (EACCES);
      return (void *) -1;
    }

  InterlockedIncrement (&shm->shm_nattch);
  _shmattach *attachnode = new _shmattach;

  attachnode->data = rv;
  attachnode->next =
    (_shmattach *) InterlockedExchangePointer ((LONG *) & shm->attachhead,
					       (long int) attachnode);


  return rv;
}

/* FIXME: evaluate getuid() and getgid() against the requested mode. Then
 * choose PAGE_READWRITE | PAGE_READONLY and FILE_MAP_WRITE  |  FILE_MAP_READ
 * appropriately
 */

/* Test result from openbsd: shm ids are persistent cross process if a handle is left
 * open. This could lead to resource starvation: we're not copying that behaviour 
 * unless we have to. (It will involve acygwin1.dll gloal shared list :[ ).
 */
/* FIXME: shmid should be a verifyable object
 */

/* FIXME: on NT we should check everything against the SD. On 95 we just emulate.
 */
extern "C" int
shmget (key_t key, size_t size, int shmflg)
{
  DWORD sd_size = 4096;
  char sd_buf[4096];
  PSECURITY_DESCRIPTOR psd = (PSECURITY_DESCRIPTOR) sd_buf;
  /* create a sd for our open requests based on shmflag & 0x01ff */
  psd = alloc_sd (getuid (), getgid (), cygheap->user.logsrv (),
		  shmflg & 0x01ff, psd, &sd_size);

  SECURITY_ATTRIBUTES sa;
  sa.nLength = sizeof (sa);
  sa.lpSecurityDescriptor = psd;
  sa.bInheritHandle = TRUE;

  char *shmname = NULL, *shmaname = NULL;
  char stringbuf[29], stringbuf1[29];

  /* FIXME: enter the checking for existing keys mutex. This mutex _must_ be system wide
   * to prevent races on shmget.
   */

  if (key != IPC_PRIVATE)
    {
      /* walk the list of currently open keys and return the id if found
       */
      shmnode *tempnode = user_data->threadinterface->shm_head;
      while (tempnode)
	{
	  if (tempnode->key == key
	      && key != IPC_PRIVATE)
	    {
	      // FIXME: free the mutex
	      if (size
		  && tempnode->shmds->shm_segsz < size)
		{
		  set_errno(EINVAL);
		  return -1;
		}
	      if ((shmflg & IPC_CREAT)
		  && (shmflg & IPC_EXCL))
		{
		  set_errno (EEXIST);
		  // FIXME: free the mutex
		  return -1;
		}
	      // FIXME: do we need to other tests of the requested mode with the
	      // tempnode->shmid mode ? testcase on unix needed.
	      // FIXME do we need a security test? We are only examining the keys we already have open.
	      // FIXME: what are the sec implications for fork () if we don't check here?
	      return tempnode->shm_id;
	    }
	  tempnode = tempnode->next;
	}
      /* couldn't find a currently open shm control area for the key.
       * Allocate a new control block - this has to be handled by the daemon */
      client_request_shm_get *req =
	new client_request_shm_get (key, size, shmflg, sd_buf);

      int rc;
      if ((rc = cygserver_request (req)))
	{
	  delete req;
	  set_errno (ENOSYS); /* daemon failed */
	  return -1;
	}

      /* we've got the id, now we open the memory area ourselves.
       * This tests security automagically
       * FIXME: make this a method of shmnode ?
       */
	snprintf (stringbuf, 29, "CYGWINSHMK0x%0qx", key);
	shmname = stringbuf;
	snprintf (stringbuf1, 29, "CYGWINSHMD0x%0qx", key);
	shmaname = stringbuf1;
/* This may need sh_none... it's only a control structure */
	HANDLE filemap = CreateFileMapping (INVALID_HANDLE_VALUE,	// system pagefile.
					    &sa,
					    PAGE_READWRITE,	// protection  
					    0x00000000,
					    getsystemallocgranularity (),
					    shmname	// object name
	  );
if (filemap != NULL)
	  {
	    /* found and successfully opened the existing mapping */
	    /* did we create it ? */
	    int oldmapping = GetLastError () == ERROR_ALREADY_EXISTS;
	    if (!oldmapping)
	      {
		/* we created a new mapping  - the daemon has failed miserably
		 * AND we didn't detect that. */
		    CloseHandle(filemap);
		    /* FIXME free mutex */
		    set_errno (ENOSYS);
		delete req;
		    return -1;
	      }

	    void *mapptr = MapViewOfFile (filemap, FILE_MAP_WRITE, 0, 0, 0);

	    if (!mapptr)
	      {
		CloseHandle(filemap);
		//FIXME: close filemap and free the mutex
		/* we couldn't access the mapped area with the requested permissions */
		set_errno (EACCES);
		delete req;
		return -1;
	      }
	
	    /* Now get the user data */
     HANDLE attachmap = CreateFileMapping (INVALID_HANDLE_VALUE, // system pagefile
                                            &sa,
                                            PAGE_READWRITE, // protection (FIXME)
                                            0x00000000,
                                            size+ size % getsystemallocgranularity (),
                                            shmaname     // object name
          );
	if (attachmap ==NULL)
	{
	system_printf ("failed to get shm attachmap\n");
	set_errno(ENOMEM);
	UnmapViewOfFile(mapptr);
CloseHandle(filemap);
	delete req;
	/*exit the mutex */
	return -1;
	}

	    shmid_ds *shmtemp = new shmid_ds;
	    if (!shmtemp)
	      {
		system_printf ("failed to malloc shm node\n");
		set_errno (ENOMEM);
		UnmapViewOfFile(mapptr);
		CloseHandle(filemap);
		CloseHandle(attachmap);
		delete req;
		/* exit mutex */
		return -1;
	      }

	    /* get the node data */
	*shmtemp = *(shmid_ds *) mapptr;

	    shmtemp->filemap = filemap;
	    shmtemp->attachmap = attachmap;
	    shmtemp->mapptr = mapptr;

    /* no need for InterlockedExchange here, we're serialised by the global mutex */
	    shmnode *tempnode = new shmnode;
	    tempnode->shmds = shmtemp;
	    tempnode->shm_id = req->parameters.shm_id;
	    tempnode->key = key;
	    tempnode->next = user_data->threadinterface->shm_head;
	    user_data->threadinterface->shm_head = tempnode;

	    /* FIXME: leave the system wide shm mutex */
	    delete req;

	    return tempnode->shm_id;
	  }

	/* We failed to open the filemapping ? */
	system_printf ("failed to open file mapping: %d\n", GetLastError ());
	// free the mutex
	// we can assume that it exists, and that it was an access problem.
	delete req;
	set_errno (EACCES);
	return -1;

    }

  /* key==IPC_PRIVATE */
  if (size < getsystemallocgranularity ())
    {
      // FIXME: leave the mutex
      set_errno (EINVAL);
      return -1;
    }

  HANDLE filemap = CreateFileMapping (INVALID_HANDLE_VALUE,	// system pagefile.
				      &sa,
				      PAGE_READWRITE,	// protection
				      0x00000000,
				      /* FIXME: round up to mallocgranularity */
				      size + getsystemallocgranularity (),
				      shmname	// object name
    );


  if (filemap == NULL)
    {
      system_printf ("failed to open file mapping: %d\n", GetLastError ());
      set_errno (geterrno_from_win_error (GetLastError (), EACCES));
      return -1;
    }

  void *mapptr = MapViewOfFile (filemap, FILE_MAP_WRITE, 0, 0, 0);

  /* Now get the user data */
  HANDLE attachmap = CreateFileMapping (INVALID_HANDLE_VALUE,	// system pagefile
					&sa,
					PAGE_READWRITE,	// protection (FIXME)
					0x00000000,
					size +
					size % getsystemallocgranularity (),
					shmaname	// object name
    );
  if (attachmap == NULL)
    {
      system_printf ("failed to get shm attachmap\n");
      set_errno (ENOMEM);
      UnmapViewOfFile (mapptr);
      CloseHandle (filemap);
      /*exit the mutex */
      return -1;
    }

  shmid_ds *shmtemp = new shmid_ds;
  if (!shmtemp)
    {
      system_printf ("failed to malloc shm node\n");
      set_errno (ENOMEM);
      UnmapViewOfFile (mapptr);
      CloseHandle (filemap);
      CloseHandle (attachmap);
      /* exit mutex */
      return -1;
    }

  /* fill out the node data */
  shmtemp->shm_perm.cuid = getuid ();
  shmtemp->shm_perm.uid = shmtemp->shm_perm.cuid;
  shmtemp->shm_perm.cgid = getgid ();
  shmtemp->shm_perm.gid = shmtemp->shm_perm.cgid;
  shmtemp->shm_perm.mode = shmflg & 0x01ff;
  shmtemp->shm_lpid = 0;
  shmtemp->shm_nattch = 0;
  shmtemp->shm_atime = 0;
  shmtemp->shm_dtime = 0;
  shmtemp->shm_ctime = time (NULL);
  shmtemp->shm_segsz = size;
  *(shmid_ds *) mapptr = *shmtemp;
  shmtemp->filemap = filemap;
  shmtemp->attachmap = attachmap;
  shmtemp->mapptr = mapptr;

  /* no need for InterlockedExchange here, we're serialised by the global mutex */
  shmnode *tempnode = new shmnode;
  tempnode->shmds = shmtemp;
  tempnode->key = key;
  tempnode->next = user_data->threadinterface->shm_head;
  tempnode->shm_id = (int) shmtemp; /* FIXME: have the daemon allocate the ID. */
  user_data->threadinterface->shm_head = tempnode;

  /* FIXME: leave the system wide shm mutex */

  return tempnode->shm_id;

}
